
# lexer.py

from __future__ import print_function

import os

import subprocess
from gcc_options import gcc_options, pkg_options

# --------------------------------------------------------------------------

fileNames = [ ] # list of file names (indexed from 1)

def indexToFileName (inx) :
    global fileNames
    if inx == 0 :
       return ""
    else :
       return fileNames [inx - 1]

def fileNameToIndex (name) :
    global fileNames
    if not name in fileNames :
       fileNames.append (name)
    return fileNames.index (name) + 1

# --------------------------------------------------------------------------

class Separators (object) :
   def __init__ (self) :
       self.value = -1  # -1 ... substring is not a valid symbol
                        # value >= 0 ... token kind
       self.items = { } # another symbols with same prefix
                        # dictionary of Separators ( keys ... first character )

class Define (object) :
   def __init__ (self) :
       self.name = ""
       self.params = None # None ... without paramaters, [ ] ... with empty parameter list
       self.dots = False
       self.code = ""
       self.fileInx = 0
       self.lineNum = 1

class CondSection (object) :
   def __init__ (self) :
      self.orig_use_input = False # outer use_input
      self.allow_reading = True # some sub-section can be used (for reading tokens)
      self.allow_else = True # "else" sub-section is possible

class LexerException (Exception) :
   pass

# --------------------------------------------------------------------------

class LexerInput (object) :
   def __init__ (self) :
       self.source = ""

       self.fileInx = 0 # file index
       self.lineNum = 1 # line number (from 1)
       self.colNum  = 1 # column number (from 1)
       self.byteOfs = 0 # byte offset (from 0)

       self.isInclude = False
       self.isMacro = False

       self.macros = { } # local macro definitions

       self.above = None # above LexerInput

       self.guard = 1 # detect ifndef ... endif
       # 1 ... before ifndef
       # 2 ... inside ifndef ... endif
       # 3 ... after endif
       self.guard_symbol = None # symbol after ifndef

       self.incl_limit = 0 # max number of nested includes (0 ... no limit)

   def readChar (self) :
       if self.byteOfs < self.sourceLen :
          self.ch = self.source [self.byteOfs]
          self.byteOfs = self.byteOfs + 1
          if self.ch == '\n' :
             self.lineNum = self.lineNum + 1
             self.colNum = 0
          elif self.ch != '\r' :
             self.colNum = self.colNum + 1
       else :
          self.ch = '\0'

   def setFileName (self, name) :
       self.fileInx = fileNameToIndex (name)

   def getFileName (self) :
       return indexToFileName (self.fileInx)

   def openFile (self, fileName) :
       fileName = os.path.abspath (fileName)
       self.setFileName (fileName)
       f = open (fileName, "r")
       self.source = f.read ()
       self.sourceLen = len (self.source)

   def openString (self, sourceText) :
       self.source = str (sourceText) # optionaly conversion from QString
       self.sourceLen = len (self.source)

   def getSource (self) :
       return self.source # !?

# --------------------------------------------------------------------------

class Lexer (object) :
   eos = 0
   identifier = 1
   number = 2
   real_number = 3
   character_literal = 4
   string_literal = 5
   separator = 6
   end_of_line = 7

   def __init__ (self) :
       self.pascal = False
       self.lexer_keywords = None # keyword table
       self.lexer_separators = None # separator table

       self.at_keywords = False # allow identifiers stated with @
       self.debug = False
       self.follow = None # object for reporting
       self.support = None # editor support

       self.reset ()

   def reset (self) :

       self.charFileInx = 0
       self.charLineNum = 1
       self.charColNum = 1
       self.charByteOfs = 0 # character (in self.ch)

       self.tokenFileInx = 0
       self.tokenLineNum = 1
       self.tokenColNum = 1
       self.tokenByteOfs = 0 # token (in self.token)

       self.prevEndOfs = 0
       self.prevFileInx = 0
       self.prevLineNum = 1
       self.prevColNum = 1
       self.prevByteOfs = 0 # previous token

       self.ch = '\0'
       self.token = self.eos
       self.tokenText = ""

       self.input = None
       self.input_stack = [ ]
       self.file_input = None # main file or include file

       self.inside_directive = False
       self.initPreproc ()

   # -----------------------------------------------------------------------

   def openFile (self, fileName) :
       # self.reset ()
       self.input = LexerInput ()
       self.input.openFile (fileName)
       self.file_input = self.input
       if self.follow : self.follow.start (self)
       self.nextChar ()
       self.nextToken ()

   def openString (self, sourceText) :
       # self.reset ()
       self.input = LexerInput ()
       self.input.openString (sourceText)
       self.file_input = self.input
       if self.follow : self.follow.start (self)
       self.nextChar ()
       self.nextToken ()

   def close (self) :
       pass

   # -----------------------------------------------------------------------

   def getFileName (self) :
       return self.file_input.getFileName () # !?

   def setFileName (self, name) :
       self.file_input.setFileName (name) # !?

   def getSource (self) :
       return self.file_input.source # !?

   def getFileInput (self) :
       inp = self.input
       while inp != None and inp.isMacro :
          inp = inp.above
       return inp

   # -----------------------------------------------------------------------

   def getPosition (self) :
       inp = self.file_input
       if inp == None :
          txt = " : : "
       else :
          txt = inp.getFileName () + ":" + str (inp.lineNum) + ":" + str (inp.colNum)
       return txt

   def info (self, text) :
       print (self.getPosition () + ": info: " + text)

   def warning (self, text) :
       print (self.getPosition () + ": warning: " + text)

   def error (self, text) :
       # print (self.getPosition () + ": error: " + text)
       raise LexerException (self.getPosition () + ": error: " + text + ", tokenText=" + self.tokenText)

   # -----------------------------------------------------------------------

   def nextChar (self) :
       self.charFileInx = self.input.fileInx
       self.charLineNum = self.input.lineNum
       self.charColNum  = self.input.colNum
       self.charByteOfs = self.input.byteOfs

       self.input.readChar ()
       self.ch = self.input.ch

       if 0 and self.follow : self.follow.oneCharacter ()
       if 0 and self.debug: print ("CHAR", self.quoteString (self.ch))

   def repeatChar (self) :
       self.charFileInx = self.input.fileInx
       self.charLineNum = self.input.lineNum
       self.charColNum  = self.input.colNum
       self.charByteOfs = self.input.byteOfs

       self.ch = self.input.ch

   # -----------------------------------------------------------------------

   def mark (self) :
       # print ("mark tokenText=" + self.tokenText)
       return (self.tokenFileInx, self.tokenLineNum, self.tokenColNum, self.tokenByteOfs)

   def rewind (self, params) :
       # print ("before rewind tokenText=" + self.tokenText)
       self.input.fileInx = params [0] # !?
       self.input.lineNum = params [1]
       self.input.colNum = params [2]
       self.input.byteOfs = params [3]
       self.nextChar ()
       self.nextToken ()
       # print ("rewind tokenText=" + self.tokenText)

   # -----------------------------------------------------------------------

   def isLetter (self, c) :
       return c >= 'A' and c <= 'Z' or c >= 'a' and c <= 'z' or c == '_'

   def isDigit (self, c) :
       return c >= '0' and c <= '9'

   def isOctDigit (self, c) :
       return c >= '0' and c <= '7'

   def isHexDigit (self, c) :
       return c >= 'A' and c <= 'F' or c >= 'a' and c <= 'f' or c >= '0' and c <= '9'

   def isLetterOrDigit (self, c) :
       return self.isLetter (c) or self.isDigit (c)

   # -----------------------------------------------------------------------

   def comment_eol (self) :
       if 0 and self.follow : self.follow.startComment (2)

       while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' :
          self.nextChar ()

       if 0 and self.follow: self.follow.comment ()

   def comment (self, mark1) :
       if 0 and self.follow : self.follow.startComment (1)

       while self.ch != '\0' and self.ch != mark1 :
          if 0 :
             self.tokenText = self.tokenText + self.ch
          self.nextChar ()
       if self.ch == mark1 :
          self.nextChar () # skip mark1
       else :
          self.error ("Unterminated comment")

       if 0 and self.follow: self.follow.comment ()

   def comment2 (self, mark1, mark2) :
       if 0 and self.follow : self.follow.startComment (2)

       prev = ' '
       while self.ch != '\0' and not (prev == mark1 and self.ch == mark2) :
          prev = self.ch
          self.nextChar ()
       if self.ch == mark2 :
          self.nextChar () # skip mark2
       else :
          self.error ("Unterminated comment")

       if 0 and self.follow : self.follow.comment ()

   def checkDigit (self) :
       if not self.isDigit (self.ch) :
          self.error ("Digit expected")

   def digits (self) :
       while self.isDigit (self.ch) :
          self.tokenText = self.tokenText + self.ch
          self.nextChar ()

   def stringChar (self) :
       if self.ch != '\\' :
          c = self.ch
          self.nextChar ()
          return c
       else :
          self.nextChar () # skip backslash

          if self.ch >= '0' and self.ch <= '7' :
             cnt = 1
             c = 0
             while self.ch >= '0' and self.ch <= '7' and cnt <= 3 :
                c = 8 * c  + ord (self.ch) - ord ('0')
                cnt = cnt + 1
                self.nextChar ()
             return chr (c)

          elif self.ch == 'x' or self.ch <= 'X' :
             self.nextChar ()
             c = 0
             while self.ch >= '0' and self.ch <= '9' or self.ch >= 'A' and self.ch <= 'Z' or self.ch >= 'a' and self.ch <= 'z':
                if self.ch >= '0' and self.ch <= '9' :
                   n = ord (self.ch) - ord ('0')
                elif self.ch >= 'A' and self.ch <= 'Z' :
                   n = ord (self.ch) - ord ('A') + 10
                else :
                   n = ord (self.ch) - ord ('a') + 10
                c = 16 * c  + n
                self.nextChar ()
             return chr (c % 255)

          elif self.ch == 'a' :
             c = '\a'
          elif self.ch == 'b' :
             c =  '\b'
          elif self.ch == 'f' :
             c =  '\f'
          elif self.ch == 'n' :
             c =  '\n'
          elif self.ch == 'r' :
             c =  '\r'
          elif self.ch == 't' :
             c =  '\t'
          elif self.ch == 'v' :
             c =  '\v'
          elif self.ch == "'" or self.ch == '"'  or ch == '?' :
             c = self.ch
          else :
             c = self.ch # same

          self.nextChar ()
          return c

   def quoteString (self, s, quote = '"') :
       result = ""
       for c in s :
          if c < ' ' :
             if c == '\a' :   result += "\\a"
             elif c == '\b' : result += "\\b"
             elif c == '\f' : result += "\\f"
             elif c == '\n' : result += "\\n"
             elif c == '\r' : result += "\\r"
             elif c == '\t' : result += "\\t"
             elif c == '\v' : result += "\\v"
             elif c == '\0' : result += "\\0"
             else :
                result += "\\x" + "{:02x}".format (ord (c))
          elif c == '\"' :
             result += "\\\""
          elif c == '\''  :
             result += "\\\'"
          else :
             result += c
       return quote + result + quote

   def lookupKeyword (self) :
       if self.lexer_keywords != None :
          if self.tokenText in self.lexer_keywords :
             self.token = self.lexer_keywords [self.tokenText]

   def processSeparator (self) :
       # print ("separator start", self.tokenText)
       if not self.inside_directive :

          # self.tokenText ... first character
          if self.lexer_separators != None :
             cont = None
             if self.tokenText in self.lexer_separators :
                cont = self.lexer_separators [self.tokenText]

             stop = False
             while not stop :
                if cont != None and self.ch in cont.items :
                   cont = cont.items [self.ch] # next map of separators
                   self.tokenText = self.tokenText + self.ch # add one character
                   # print ("separator character", self.ch)
                   self.nextChar ()
                elif cont != None and cont.value >= 0 :
                   # print ("separator", self.tokenText, "separator value ", cont.value)
                   self.token = cont.value # new separator
                   stop = True
                else :
                   self.error ("Unknown separator")

   def backStep (self) :
       self.tokenColNum = self.tokenColNum - 1
       self.tokenByteOfs = self.tokenByteOfs - 1

   # -----------------------------------------------------------------------

   def normalToken (self) :
       self.token = self.eos
       self.tokenText = ""

       slash = False
       stop = False
       whiteSpace = True

       while whiteSpace and self.ch != '\0' and not stop:
          if self.inside_directive :
             while (self.ch != '\0' and self.ch <= ' ') or self.ch == '\\' :
                if self.ch == '\\' :
                   if 0 and self.follow: self.follow.whiteSpace ()
                   self.nextChar ()
                   if self.ch == '\r' :
                      if 0 and self.follow: self.follow.whiteSpace ()
                      self.nextChar ()
                   if self.ch == '\n' :
                      if 0 and self.follow: self.follow.whiteSpace ()
                      self.nextChar ()
                else :
                   if self.ch == '\r' or self.ch == '\n' :
                      stop = True
                   if 0 and self.follow: self.follow.whiteSpace ()
                   self.nextChar ()
          else :
             while self.ch != '\0' and self.ch <= ' ' :
                if 0 and self.follow: self.follow.whiteSpace ()
                self.nextChar ()

          whiteSpace = False
          if self.ch == '/' :
             self.nextChar () # skip '/'
             if self.ch == '/' :
                self.nextChar () # skip '/'
                self.comment_eol ()
                whiteSpace = True # check again for white space
             elif self.ch == '*' :
                self.nextChar () # skip '*'
                self.comment2 ('*', '/')
                whiteSpace = True # check again for white space
             else:
                slash = True # produce '/' token

       self.tokenFileInx = self.charFileInx
       self.tokenLineNum = self.charLineNum
       self.tokenColNum = self.charColNum
       self.tokenByteOfs = self.charByteOfs

       if slash :
          self.token = self.separator
          self.tokenText = '/'
          self.backStep ()
          self.processSeparator ()

       elif stop :
          self.token = self.end_of_line

       elif self.ch == '\0' :
          self.token = self.eos

       elif self.isLetter (self.ch) :
          self.token = self.identifier
          while self.isLetterOrDigit (self.ch) :
             self.tokenText = self.tokenText + self.ch
             self.nextChar ()
          # NO self.lookupKeyword ()

       elif self.ch == '@' and self.at_keywords :
          self.token = self.identifier
          while self.isLetterOrDigit (self.ch) or self.ch == '@':
             self.tokenText = self.tokenText + self.ch
             self.nextChar ()
          # NO self.lookupKeyword ()

       elif self.isDigit (self.ch) :
          self.token = self.number
          cont = True
          if self.ch == "0" :
             self.tokenText = self.tokenText + self.ch
             self.nextChar ()
             if self.ch == "x" :
                self.tokenText = self.tokenText + self.ch
                self.nextChar ()
                while self.isHexDigit (self.ch) :
                   self.tokenText = self.tokenText + self.ch
                   self.nextChar ()
                cont = False
             else :
                while self.isOctDigit (self.ch) :
                   self.tokenText = self.tokenText + self.ch
                   self.nextChar ()
                cont = False
          if cont :
             self.digits ()
             if self.ch == '.' :
                self.token = self.real_number
                self.tokenText = self.tokenText + self.ch # store '.'
                self.nextChar () # skip '.'
                # NO self.checkDigit ()
                self.digits ()
             if self.ch == 'e' or self.ch == 'E':
                self.token = self.real_number
                self.tokenText = self.tokenText + self.ch # store 'e'
                self.nextChar () # skip 'e'
                if self.ch == '+' or self.ch == '-':
                   self.tokenText = self.tokenText + self.ch # store '+' or '-'
                   self.nextChar () # skip '+' or '-'
                self.checkDigit ()
                self.digits ()
             if self.token == self.real_number :
                if self.ch == 'f' or self.ch == 'F':
                   self.nextChar ()
             else :
                if self.ch == 'l' or self.ch == 'L':
                   self.nextChar ()

       elif self.ch == '\'' :
          self.token = self.character_literal
          self.nextChar ()
          while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' and self.ch != '\'' :
             self.tokenText = self.tokenText + self.stringChar ()
          if self.ch != '\'' :
             self.error ("Unterminated string")
          self.nextChar ()

       elif self.ch == '\"' :
          self.token = self.string_literal
          self.nextChar ()
          while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' and self.ch != '\"' :
             self.tokenText = self.tokenText + self.stringChar ()
          if self.ch != '\"' :
             self.error ("Unterminated string")
          self.nextChar ()

       else :
          self.token = self.separator
          self.tokenText = self.ch
          first_ch = self.ch
          self.nextChar ()
          if first_ch != '#' :
             self.processSeparator ()

   # -----------------------------------------------------------------------

   def pascalSkipSpaces (self) :
       while self.ch != '\0' and self.ch <= ' ' :
          self.nextChar ()

   def pascalReadName (self) :
       self.pascalSkipSpaces ()
       name = ""
       while self.isLetterOrDigit (self.ch) :
          name = name + self.ch.lower ()
          self.nextChar ()
       if name == "" :
          self.error ("Identifier expected")
       return name

   def pascalReadFileName (self, term) :
       self.pascalSkipSpaces ()
       name = ""
       while self.ch != '\0' and self.ch > ' ' and self.ch != term :
          name = name + self.ch
          self.nextChar ()
       if name == "" :
          self.error ("Include file name expected")
       return name

   def pascalDirective (self, term) :
       self.pascalSkipSpaces ()
       if self.ch == '$' :
          self.nextChar ()
          name = self.pascalReadName ()
          if name == "include" :
             fileName = self.pascalReadFileName ()
             # !? self.simpleInclude (fileName)
          elif name == "define" :
             id = self.pascalReadName ()
             self.addMacro (id, "")
          elif name == "undef" :
             id = self.pascalReadName ()
             self.undefMacro (id)
          elif name == "ifdef" :
             id = self.pascalReadName ()
             self.simpleIf (id in self.macros)
          elif name == "ifndef" :
             id = self.pascalReadName ()
             self.simpleIf (id not in self.macros)
          elif name == "else" :
             self.simpleElse ()
          elif name == "endif" :
             self.simpleEndif ()
          else :
             self.error ("Unknown directive")
          self.pascalSkipSpaces ()
          if self.ch != term :
             self.error ("End of directive expected")

   def pascalToken (self) :
       self.token = self.eos
       self.tokenText = ""

       whiteSpace = True
       slash = False
       parenthesis = False

       while whiteSpace and self.ch != '\0':
          whiteSpace = False
          while self.ch != '\0' and self.ch <= ' ' :
             self.nextChar ()
          if self.ch == '/' :
             self.nextChar () # skip '/'
             if self.ch == '/' :
                self.nextChar () # skip '/'
                self.comment_eol ()
                whiteSpace = True # check again for white space
             else:
                slash = True # produce '/' token
          if self.ch == '{' :
             self.nextChar () # skip '{'
             self.pascalDirective ('}')
             self.comment ('}')
             whiteSpace = True # check again for white space
          if self.ch == '(' :
             self.nextChar () # skip '('
             if self.ch == '*' :
                self.nextChar () # skip '*'
                self.pascalDirective ('*')
                self.comment2 ('*', ')')
                whiteSpace = True # check again for white space
             else :
                parenthesis = True # produce '(' token

       self.tokenFileInx = self.charFileInx
       self.tokenLineNum = self.charLineNum
       self.tokenColNum = self.charColNum
       self.tokenByteOfs = self.charByteOfs

       if slash :
          self.token = self.separator
          self.tokenText = '/'
          self.backStep ()
          self.processSeparator ()

       elif parenthesis :
          self.token = self.separator
          self.tokenText = '('
          self.backStep ()
          self.processSeparator ()

       elif self.ch == '\0' :
          self.token = self.eos

       elif self.isLetter (self.ch) :
          self.token = self.identifier
          while self.isLetterOrDigit (self.ch) :
             self.tokenText = self.tokenText + self.ch
             self.nextChar ()
          self.tokenText = self.tokenText.lower ()
          # NO self.lookupKeyword ()

       elif self.isDigit (self.ch) :
          self.token = self.number
          self.digits ()
          if self.ch == '.' :
             self.nextChar () # skip '.'
             if self.ch == '.' : # two dots
                self.backStep()
             else :
                self.token = self.real_number
                self.tokenText = self.tokenText + '.' # store '.'
                self.checkDigit ()
                self.digits ()
          if self.ch == 'e' or self.ch == 'E':
             self.token = self.real_number
             self.tokenText = self.tokenText + self.ch # store 'e'
             self.nextChar () # skip 'e'
             if self.ch == '+' or self.ch == '-':
                self.tokenText = self.tokenText + self.ch # store '+' or '-'
                self.nextChar () # skip '+' or '-'
             self.checkDigit ()
             self.digits ()

       elif self.ch == '\'' :
          self.token = self.string_literal
          self.nextChar ()
          again = True
          while again :
             again = False
             while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' and self.ch != '\'' :
                self.tokenText = self.tokenText + self.ch
                self.nextChar ()
             if self.ch == '\'' :
                self.nextChar () # skip quote
                if self.ch == '\'' : # two quotes
                   self.tokenText = self.tokenText + self.ch # store one quote
                   self.nextChar ()
                   again = True
             else :
                self.error ("Unterminated string")

       else :
          self.token = self.separator
          self.tokenText = self.ch
          self.nextChar ()
          self.processSeparator ()

   # -----------------------------------------------------------------------

   def initPreproc (self) :
       self.incl_path = [ "." ]

       self.macros = { }
       self.active_macros = { }

       self.use_input = True
       self.cond_stack = [ ]

       self.guard_map = { }

       self.tolerate_missing_include = False
       self.include_files = [ ]

       self.expand_limit = 100

       self.preprocDir = "_input"
       self.inputPrefix = "input_"
       self.preprocPrefix = "preproc_"

   def addInclDir (self, path) :
       self.incl_path.append (path)
       # print ("INCL DIR", path)

   def newMacro (self, name, value) :
       defn = Define ()
       defn.name = name
       defn.code = value
       defn.fileInx = self.tokenFileInx
       defn.lineNum = self.tokenLineNum
       return defn

   def addMacro (self, name, value) :
       defn = self.newMacro (name, value)
       self.macros [defn.name] = defn
       return defn

   def addMacroWithParams (self, name, params, value) :
       defn = self.addMacro (name, value)
       defn.params = [ ]
       for param in params :
          defn.params.append (param)
       return defn

   def undefMacro (self, name) :
       if name in self.macros :
          del self.macros [name]

   def searchFile (self, filename, incl_sys) :
       result = None
       if not incl_sys :
          path = os.path.dirname (self.getFileName ())
          test = os.path.join (path, filename)
          if os.path.exists (test):
             result = test
       for path in self.incl_path :
          test = os.path.join (path, filename)
          if os.path.exists (test):
             result = test
             break
       return result

   def searchNextFile (self, filename) :
       inp = self.input
       while inp != None and inp.isMacro :
          inp = inp.above

       dirName = ""
       if inp != None :
          dirName = os.path.dirname (inp.getFileName ())

       result = None
       accept = False
       for path in self.incl_path :
          if accept :
             test = os.path.join (path, filename)
             if os.path.exists (test):
                result = test
                # result = os.path.abspath (result)
                break
          else :
             if path == dirName :
                accept = True
       return result

   def simpleToken (self) :
       "token without expansion"
       self.prevFileInx = self.tokenFileInx
       self.prevLineNum = self.tokenLineNum
       self.prevColNum = self.tokenColNum
       self.prevByteOfs = self.tokenByteOfs
       self.prevEndOfs = self.charByteOfs

       if self.pascal :
          self.pascalToken ()
       else :
          self.normalToken ()

       if self.debug : print ("SIMPLE TOKEN", self.quoteString (self.tokenText), "line", str (self.tokenLineNum), "col", str (self.tokenColNum))

   def isToken (self, value) :
       return self.tokenText == value

   def verifyToken (self, value) :
       if not self.isToken (value) :
          self.error (value + " expected")

   # -----------------------------------------------------------------------

   def searchMacro (self, name) :
       defn = None
       inp = self.input
       while defn == None and inp != None :
          if name in inp.macros :
             defn = inp.macros [name]
          inp = inp.above
       if defn == None and name in self.macros :
          defn = self.macros [name]
       # if defn != None :
       #   print ("FOUND", name)
       return defn

   def expandName (self, name) :
       defn = self.searchMacro (name)
       if defn != None and defn.params == None and not defn.dots :
          return defn.code
       else :
          return name

   def skipSpaces (self) :
       while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' and self.ch <= ' ' :
          self.nextChar ()

   def expandToken (self) :
       again = True
       while again :
          "read simple token"
          self.simpleToken ()
          # if self.debug : print ("simple token", self.tokenText, self.use_input)
          again = False
          quote = False
          if self.use_input and self.input.isMacro and self.tokenText == '#' :
             quote = True
             self.simpleToken ()
          if self.use_input and self.token == self.identifier :
             "look for macro definition"
             name = self.tokenText
             if self.input.isMacro :
                self.skipSpaces ()
                if self.ch == '#' :
                   name = self.expandName (name)
                   while self.ch == '#' : # !?
                      self.nextChar () # skip hash mark
                      if self.ch != '#' :
                         # self.error ("## expected")
                         self.warning ("## expected")
                      self.nextChar () # skip second hash mark
                      self.simpleToken () # skip identifier (as token)
                      if self.token != self.identifier :
                         self.warning ("Identifier expected")
                      else :
                         name2 = self.tokenText
                         name = name + self.expandName (name2)
                         self.skipSpaces ()
                      # keep token equal to last identifier
                   if self.debug : print ("COMPOUND NAME", name)

             if quote :
                value = self.expandName (name)
                self.token = self.string_literal
                self.tokenText = value
                if self.debug : print ("STRING", name, self.quoteString (value))
             else :
                if not self.active_macros.get (name) :
                   defn = self.searchMacro (name)
                   if defn != None : #
                      "expand macro"
                      if len (self.input_stack) > self.expand_limit :
                         self.error ("Too many nested macros")
                      if self.follow : self.follow.expandMacro (defn)
                      if self.debug : print ("EXPAND MACRO", defn.name)
                      next_input = LexerInput ()
                      missing_parameters = False

                      if defn.params != None or defn.dots :
                         self.simpleToken () # skip macro name
                         if not self.isToken ('(') :
                            missing_parameters = True
                         else :
                            if self.follow: self.follow.openParameters ()
                            if self.follow : self.follow.paramToken ()
                            self.simpleToken () # skip '('
                            cnt = len (defn.params)
                            inx = 0
                            args = None
                            while inx < cnt or defn.dots and not self.isToken (')') :
                                if inx < cnt :
                                   param_name = defn.params [inx]
                                else :
                                   param_name = ""
                                if self.follow: self.follow.openParam (param_name)
                                if 0 and self.debug : print ("OPEN PARAM", param_name)
                                value = ""
                                level = 0
                                while not (level == 0 and (self.isToken (',') or self.isToken (')'))):
                                   if self.isToken ('(') :
                                      level = level + 1
                                   if self.isToken (')') :
                                      level = level - 1
                                   if value != "" :
                                      value = value + " "
                                   value = value + self.tokenText
                                   if self.follow : self.follow.paramToken ()
                                   self.simpleToken ()
                                if self.follow: self.follow.closeParam ()
                                if 0 and self.debug : print ("CLOSE PARAM", param_name, "=", self.quoteString (value))
                                if inx < cnt :
                                   param = self.newMacro (param_name, value)
                                   next_input.macros [param.name] = param
                                else :
                                   if args == None :
                                      args = self.newMacro ("__VA_ARGS__", value)
                                      next_input.macros [args.name] = args
                                   else :
                                      args.code = args.code + "," + value
                                if inx+1 < cnt or not self.isToken (')') :
                                   self.verifyToken (',')
                                   if self.follow : self.follow.paramToken ()
                                   self.simpleToken ()
                                inx = inx + 1
                            self.verifyToken (')')
                            if self.follow : self.follow.paramToken ()
                            # NO self.simpleToken ()
                            if self.follow: self.follow.closeParameters ()

                      if not missing_parameters :
                         next_input.isMacro = True
                         next_input.macro_name = defn.name
                         next_input.above = self.input
                         self.input_stack.append (self.input)
                         self.active_macros [next_input.macro_name] = True
                         self.input = next_input
                         self.input.openString (defn.code + " ")
                         self.nextChar ()
                         again = True

          if not again :
             if self.isEndOfSource () and len (self.input_stack) != 0 :
                if self.input.isMacro : # only for macros, not for include files
                   if self.follow: self.follow.closeExpand ()
                   if 0 and self.debug : print ("END EXPAND")
                   self.active_macros [self.input.macro_name] = False
                   self.input = self.input_stack.pop ()
                   self.file_input = self.getFileInput ()
                   self.repeatChar ()
                   again = True

          if not again :
             if self.debug : print ("EXPANDED TOKEN", self.tokenText, self.expandName (self.tokenText))

   # -----------------------------------------------------------------------

   def defineDirective (self) :
       self.simpleToken () # skip define

       if not self.isIdentifier () :
          self.error ("Macro name expected")
       name = self.tokenText
       defn = self.newMacro (name, "")
       with_params = (self.ch == '(') # no space before '(' allowed
       self.simpleToken () # skip name

       if with_params :
          defn.params = [ ]
          self.simpleToken () # skip '('
          while not self.isToken (')') :
             if self.isToken ("...") :
                defn.dots = True
                self.simpleToken () # skip "..."
                self.verifyToken (')')
             else :
                if not self.isIdentifier () :
                   self.error ("Parameter name expected")
                defn.params.append (self.tokenText)
                self.simpleToken ()
                if not self.isToken (')') :
                   self.verifyToken (',')
                   self.simpleToken ()
          self.simpleToken () # skip ')'

       while self.token != self.eos and self.token != self.end_of_line :
           defn.code = defn.code + self.tokenText
           if self.ch != '\0' and self.ch != '\r' and self.ch != '\n' and self.ch <= ' ' :
              defn.code = defn.code + " "
           self.simpleToken ()

       self.macros [defn.name] = defn
       if self.follow: self.follow.defineDirective (defn)
       if self.debug : print ("DEFINE", defn.name, defn.params, defn.code)
       self.info ("DEFINE " + defn.name)

   def undefDirective (self) :
       self.simpleToken () # skip undef
       if not self.isIdentifier () :
          self.error ("Macro name expected")

       name = self.tokenText
       if name in self.macros :
          del self.macros [name]
       if self.follow: self.follow.undefDirective (name)

       self.simpleToken () # skip name

   def includeFileName (self) :
       incl_sys = False
       if self.token == self.string_literal :
          result = self.tokenText
       elif self.isToken ('<') :
          incl_sys = True
          result = ""
          while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' and self.ch != '>' :
             result = result + self.ch
             self.nextChar ()
          if self.ch != '>' :
             self.error ("Missing >")
          self.nextChar ()
       else :
          self.error ("Include file name expected")
       self.simpleToken ()
       if not self.use_input :
          result = None
       return (result, incl_sys)

   # -----------------------------------------------------------------------

   def readMessage (self) :
       text = self.ch
       while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' : # characters, not tokens
          if self.ch == '\\' :
             self.nextChar ()
             ok = False
             if self.ch == '\r' :
                self.nextChar ()
                ok = True
             if self.ch == '\n' :
                self.nextChar ()
                ok = True
             if not ok :
                text = text + '\\'
          else :
             text = text + self.ch
             self.nextChar ()
       self.token = self.end_of_line # !?
       return text

   def isDefined (self) :
       if not self.isIdentifier () :
          self.error ("Identifier expected")
       name = self.tokenText
       self.simpleToken ()
       return name in self.macros

   def readValue (self) :
       again = True
       while again :
          again = False
          # print ("READ_VALUE", self.tokenText)
          if self.isIdentifier () :
             if self.tokenText == "defined" :
                self.simpleToken ()
                if self.isToken ('(') :
                   self.simpleToken ()
                   value = self.isDefined ()
                   self.verifyToken (')')
                   self.simpleToken ()
                else :
                   value = self.isDefined ()
             else :
                name = self.tokenText
                if name in self.macros :
                   again = True
                else :
                   # self.error ("Unknown identifier " + name)
                   value = 0 # !?
                self.expandToken ()
          elif self.isNumber () :
             text = self.tokenText
             if text.endswith ('l') or text.endswith ('L') :
                text = text [:-1]
             if text.startswith ("0x") :
                value = int (text, 16)
             elif text.startswith ("0") :
                value = int (text, 8)
             else :
                value = int (text)
             self.expandToken ()
          elif self.isToken ('-') :
             self.expandToken ()
             value = - self.readValue ()
          elif self.isToken ('!') :
             self.expandToken ()
             value = not self.readValue ()
          elif self.isToken ('(') :
             self.expandToken ()
             value = self.expression ()
             self.verifyToken (')')
             self.expandToken ()
          else :
             self.error ("Value expected " + self.tokenText)
       return value

   def readArit (self) :
       func = ""
       prio = 0
       if self.isToken ('+') or self.isToken ('-') :
          func = self.tokenText
          prio = 11
       elif self.isToken ('*') or self.isToken ('/') or self.isToken ('%') :
          func = self.tokenText
          prio = 12
       elif self.isToken ('&') :
          if self.ch == '&' :
             self.nextChar ()
             func = "&&"
             prio = 4
          else :
             func = self.tokenText
             prio = 7
       elif self.isToken ('|') :
          if self.ch == '|' :
             self.nextChar ()
             func = "||"
             prio = 3
          else :
             func = self.tokenText
             prio = 5
       elif self.isToken ('<') :
          if self.ch == '<' :
             self.nextChar ()
             func = "<<"
             prio = 10
          elif self.ch == '=' :
             self.nextChar ()
             func = "<="
             prio = 9
          else :
             func = self.tokenText
             prio = 9
       elif self.isToken ('>') :
          if self.ch == '>' :
             self.nextChar ()
             func = ">>"
             prio = 10
          elif self.ch == '=' :
             self.nextChar ()
             func = ">="
             prio = 9
          else :
             func = self.tokenText
             prio = 9
       elif self.isToken ('=') :
          if self.ch == '=' :
             self.nextChar ()
             func = "=="
             prio = 8
          else :
             self.error ("Unknown operator")
       elif self.isToken ('!') :
          if self.ch == '=' :
             self.nextChar ()
             func = "!="
             prio = 8
          else :
             self.error ("Unknown operator")

       elif self.isToken ("&&") :
          func = self.tokenText
          prio = 4
       elif self.isToken ("||") :
          func = self.tokenText
          prio = 3
       elif self.isToken ("<<") or self.isToken (">>"):
          func = self.tokenText
          prio = 10
       elif self.isToken ("<=") or self.isToken (">="):
          func = self.tokenText
          prio = 9
       elif self.isToken ("==") or self.isToken ("!=") :
          func = self.tokenText
          prio = 8

       else :
          self.error ("Unknown operator " + self.tokenText)
       self.expandToken ()
       return (func, prio)

   def calculate (self, a, func, b) :
       # if func == ">=" :
       #    print (">=", a, b, a >= b)
       # if func == "&&" :
       #    print ("&&", a, b, bool (a) and bool (b))
       #    self.info ("&&")
       # if func == "||" :
       #    print ("||", a, b, bool (a) or bool (b))
       #    self.info ("||")
       if func == "*" : return a * b
       elif func == "/" : return a / b
       elif func == "%" : return a % b
       elif func == "+" : return a + b
       elif func == "-" : return a - b
       elif func == "<<" : return a << b
       elif func == ">>" : return a >> b
       elif func == "<" : return a < b
       elif func == "<=" : return a <= b
       elif func == ">" : return a > b
       elif func == ">=" : return a >= b
       elif func == "==" : return a == b
       elif func == "!=" : return a != b
       elif func == "^" : return a ^ b
       elif func == "&" : return a & b
       elif func == "|" : return a | b
       elif func == "&&" : return bool (a) and bool (b)
       elif func == "||" : return bool (a) or bool (b)
       elif func == "," : return b
       else: self.error ("unknown operator " + func)

   def simpleExpression (self) :
       values = [ ]
       functions = [ ]
       priorities = [ ]

       value = self.readValue ()
       # if self.debug : print ("VALUE", value)
       values.append (value)

       while self.token != self.eos and self.token != self.end_of_line and not self.isToken (')') and not self.isToken ('?') and not self.isToken (':')  :

          next_func, next_prio = self.readArit ()
          # if self.debug : print ("NEXT FUNC", next_func)

          while len (priorities) > 0 and priorities [-1] >= next_prio :
             right = values.pop ()
             left = values.pop ()
             func = functions.pop ()
             priorities.pop ()
             value = self.calculate (left, func, right)
             # if self.debug : print ("CALC", left, func, right, "->", value)
             values.append (value)

          functions.append (next_func)
          priorities.append (next_prio)

          value = self.readValue ()
          # if self.debug : print ("VALUE", value)
          values.append (value)

       value = values.pop ()
       while len (priorities) > 0:
          left = values.pop ()
          func = functions.pop ()
          priorities.pop ()
          value = self.calculate (left, func, value)

       return value

   def expression (self) :
       if self.debug : print ("begin expr")
       value = self.simpleExpression ()
       if self.isToken ('?') :
          self.expandToken ()
          first = self.simpleExpression ()
          self.verifyToken (':')
          self.expandToken ()
          second = self.simpleExpression ()
          if value :
             value = first
          else :
             value = second
       if self.debug : print ("end expr", value)
       return value

   # -----------------------------------------------------------------------

   def simpleIf (self, cond) :
       section = CondSection ()
       section.orig_use_input = self.use_input
       section.allow_reading = self.use_input
       section.allow_else = True
       self.cond_stack.append (section)
       if self.use_input :
          self.use_input = cond
          if cond :
             section.allow_reading = False # skip following sub-sections

   def simpleElif (self, cond) :
       if len (self.cond_stack) == 0 :
          self.error ("Unexpected elif")
       section = self.cond_stack [-1]
       if not section.allow_else :
          self.error ("Unexpected elif after else")

       self.use_input = section.allow_reading # important
       if self.use_input :
          self.use_input = cond
          if cond :
             section.allow_reading = False # skip following sub-sections

   def simpleElse (self) :
       if len (self.cond_stack) == 0 :
          self.error ("Unexpected else")
       section = self.cond_stack [-1]
       if not section.allow_else :
          self.error ("Duplicated else")

       self.use_input = section.allow_reading # important
       if self.use_input :
          section.allow_reading = False # skip following sub-sections

       section.allow_else = False

   def simpleEndif (self) :
       if len (self.cond_stack) == 0 :
          self.error ("Unexpected endif")
       section = self.cond_stack.pop ()
       self.use_input = section.orig_use_input

   # -----------------------------------------------------------------------

   def ifDirective (self) :
       self.expandToken ()
       if self.use_input :
          cond = bool (self.expression ())
       else :
          cond = False
          self.readMessage ()
       self.simpleIf (cond)
       # self.info ("if " + str (cond) + " -- " + str (len (self.cond_stack)))

   def ifdefDirective (self, negation = False) :
       self.simpleToken ()
       cond_name = self.tokenText
       if self.use_input :
          cond = self.isDefined ()
          if negation :
             cond = not cond
       else :
          cond = False
          self.readMessage ()
       self.simpleIf (cond)
       # self.info ("ifdef " + str (cond) + " -- " + str (len (self.cond_stack)))
       return cond_name

   def elifDirective (self) :
       if len (self.cond_stack) == 0 :
          self.error ("Unexpected else")
       section = self.cond_stack [-1]

       self.use_input = section.allow_reading # important, before expandToken
       self.expandToken ()

       if self.use_input : # important
          cond = bool (self.expression ())
       else :
          cond = False
          self.readMessage ()
       # self.info ("ELIF " + str (cond))
       self.simpleElif (cond)
       # self.info ("elif " + str (cond) + " -- " + str (len (self.cond_stack)))

   def elseDirective (self) :
       self.simpleToken ()
       self.simpleElse ()
       # self.info ("else " + " -- " + str (len (self.cond_stack)))

   def endifDirective (self) :
       # self.info ("endif " + " -- " + str (len (self.cond_stack)))
       self.simpleToken ()
       self.simpleEndif ()

   # -----------------------------------------------------------------------

   def readOptions (self, options) :
       for line in options :
           words = line.split (" ", 1)
           if words [0] == "-I" :
              path = words [1]
              # print ("include", path)
              self.addInclDir (path)
           elif words [0] == "-D" :
              params = words[1].split ('=')
              name = params [0]
              value = params [1]
              # print ("define", name, value)
              self.addMacro (name, value)
           elif line.startswith ("-I") :
              path = line [2:]
              # print ("include", path)
              self.addInclDir (path)
           elif line.startswith ("-D") :
              line = line [2:]
              if line.find ("=") >= 0 :
                 params = line.split ('=')
                 name = params [0]
                 value = params [1]
              else :
                 name = line
                 value = ""
              # print ("define", name, value)
              self.addMacro (name, value)

   # -----------------------------------------------------------------------

   def make_dir (self, path) :
       if not os.path.isdir (path) :
          self.makedir (os.path.dirName (path))
          os.mkdir (path)

   def preprocess (self) :
       sourceFileName = self.getFileName ()

       inputFileName = os.path.join (self.preprocDir,
                                     self.inputPrefix + os.path.basename (sourceFileName))

       self.make_dir (os.path.dirname (inputFileName))

       text = ""
       while self.ch != '\0' :
          text += self.ch
          self.nextChar ()

       g = open (inputFileName, "w")
       g.write (text)
       g.close ()

       # run cpp command

       cmd = "cpp"

       for s in self.incl_path :
           cmd = cmd + " " + "-I" + s

       for name in self.macros :
           defn = self.macros [name]
           if defn.params == None and not defn.dots :
              cmd = cmd + " " + "-D" + defn.name
              if defn.code != "" :
                 cmd = cmd + "=" + defn.code

       cmd = cmd + " "  + inputFileName

       # print (cmd)
       f = subprocess.Popen (cmd, shell=True, stdout=subprocess.PIPE).stdout
       result = ""
       for s in f :
           if not s.startswith ('#') :
              result = result + s

       # write preprocessed text

       preprocFileName = os.path.join (self.preprocDir,
                                       self.preprocPrefix + os.path.basename (sourceFileName))

       self.make_dir (os.path.dirname (preprocFileName))

       g = open (preprocFileName, "w")
       g.write (result)
       g.close ()

       return preprocFileName

   # -----------------------------------------------------------------------

   def directive (self) :
       incl_file_name = None
       incl_sys = False
       incl_next = False
       incl_lim = 0

       entry_use_input = self.use_input
       cond_directive = False
       cond_name = ""
       self.inside_directive = True

       if self.follow: self.follow.startDirective ()
       self.simpleToken () # skip hash mark
       directive_name = self.tokenText

       if self.isToken ("if") :
          self.ifDirective ()
          cond_directive = True
          if self.follow : self.follow.ifDirective ()

       elif self.isToken ("ifdef") :
          self.ifdefDirective ()
          cond_directive = True
          if self.follow : self.follow.ifdefDirective ()

       elif self.isToken ("ifndef") :
          cond_name = self.ifdefDirective (negation = True)
          cond_directive = True
          if self.follow : self.follow.ifndefDirective ()

       elif self.isToken ("elif") :
          self.elifDirective ()
          cond_directive = True
          if self.follow : self.follow.elifDirective ()

       elif self.isToken ("else") :
          self.elseDirective ()
          cond_directive = True
          if self.follow : self.follow.elseDirective ()

       elif self.isToken ("endif") :
          self.endifDirective ()
          cond_directive = True
          if self.follow : self.follow.endifDirective ()

       elif not self.use_input :
          self.readMessage () # read rest of line
          if self.follow : self.follow.unusedDirective ()

       else :
          if self.isToken ("define") :
             self.defineDirective ()

          elif self.isToken ("undef") :
             self.undefDirective ()

          elif self.isToken ("include") :
             self.expandToken () # skip include
             incl_file_name, incl_sys = self.includeFileName ()

          elif self.isToken ("include_next") :
             self.expandToken ()
             incl_file_name, incl_sys = self.includeFileName ()
             incl_next = True

          elif self.isToken ("warning") :
             text = self.readMessage ()
             if self.follow : self.follow.warningDirective (text)
             self.warning (text)

          elif self.isToken ("error") :
             text = self.readMessage ()
             if self.follow : self.follow.errorDirective (text)
             self.error (text)

          else :
             pragma = False
             if self.isToken ("pragma") :
                self.simpleToken () # skip pragma
                pragma = True

             if self.isToken ("eval") :
                self.expandToken ()
                value = self.expression ()
                self.info ("EVAL " + str (value))
                if self.follow : self.follow.evalDirective (str (value))

             if self.isToken ("print") :
                self.expandToken ()
                value = ""
                while self.token != self.eos and self.token != self.end_of_line :
                   value = value + self.tokenText + " "
                   self.expandToken ()
                self.info ("PRINT " + value)
                if self.follow : self.follow.printDirective (value)

             elif self.isToken ("preproc") :
                self.simpleToken ()
                self.preprocess ()

             elif self.isToken ("include_one") :
                self.expandToken ()
                incl_file_name, incl_sys = self.includeFileName ()
                incl_lim = 1

             elif self.isToken ("include_two") :
                self.expandToken ()
                incl_file_name, incl_sys = self.includeFileName ()
                incl_lim = 2

             elif self.isToken ("include_lim") :
                self.expandToken ()
                if self.token != self.number :
                   self.error ("Number of leveles expected")
                incl_lim = int (self.tokenText) # !?
                incl_file_name, incl_sys = self.includeFileName ()

             elif self.isToken ("include_dir") :
                self.expandToken ()
                if self.token != self.string_literal :
                   self.error ("String with directory name expected")
                path = self.tokenText
                self.addInclDir (path)
                self.simpleToken ()

             elif self.isToken ("gcc_options") :
                self.readOptions (gcc_options ())
                self.simpleToken ()

             elif self.isToken ("pkg") :
                self.expandToken ()
                if self.token != self.string_literal :
                   self.error ("String with package name expected")
                name = self.tokenText
                self.readOptions (pkg_options (name))
                self.simpleToken ()

             elif self.isToken ("tolerate_missing_include") :
                self.tolerate_missing_include = True
                self.simpleToken ()

             elif self.isToken ("debug") :
                self.debug = True
                self.simpleToken ()

             else :
                if pragma :
                   self.warning ("Unknown pragma directive " + self.tokenText)
                   text = self.readMessage ()
                elif self.isIdentifier ():
                   # self.error ("Unknown directive " + self.tokenText)
                   self.warning ("Unknown directive " + self.tokenText)
                   text = self.readMessage ()
                else :
                   # self.error ("Directive name expected")
                   self.warning ("Directive name expected")
                   text = self.readMessage ()

       if self.token != self.end_of_line :
          self.warning ("End of line (after directive) expected")
          self.readMessage ()

       self.inside_directive = False
       if self.follow: self.follow.endDirective ()

       # ifndef .. endif guard

       if self.input.isInclude :
          if entry_use_input and self.input.guard == 1 and directive_name == "ifndef" :
             self.input.guard = 2
             self.input.guard_symbol = cond_name
             # print ("GUARD 2")
          elif self.use_input and self.input.guard == 2 and directive_name == "endif" :
             self.input.guard = 3
             # print ("GUARD 3")
          elif entry_use_input :
             # if self.input.guard != 0 : print ("GUARD 0 " + str (self.token) + self.tokenText, directive_name)
             self.input.guard = 0

       # open include file

       if incl_file_name != None :
          inp = self.file_input
          if inp.incl_limit == 1 :
             self.warning ("Ignored include file " + incl_file_name)
             incl_file_name = None
          elif inp.incl_limit != 0 and incl_lim == 0:
             incl_lim = inp.incl_limit - 1

       if incl_file_name != None :
          if incl_next :
             full_name = self.searchNextFile (incl_file_name)
          else :
             full_name = self.searchFile (incl_file_name, incl_sys)

          if full_name == None :
             if self.tolerate_missing_include :
                self.warning ("Include file not found " + incl_file_name)
                self.expandToken ()
             else :
                self.error ("Include file not found " + incl_file_name)

          if full_name != None :
             key = os.path.abspath (full_name)
             if key in self.guard_map :
                cond = self.guard_map [key]
                if cond in self.macros :
                   # print ("SKIP INCL ", incl_file_name)
                   full_name = None

          if full_name != None :
             self.info ("BEFORE INCL " + full_name)
             next_input = LexerInput ()
             next_input.above = self.input
             next_input.isInclude = True
             next_input.incl_limit = incl_lim
             self.input_stack.append (self.input)
             self.input = next_input
             self.input.openFile (full_name)
             self.file_input = self.input
             # if full_name in self.include_files :
                # print ("ALREADY INCLUDED ", full_name)
             self.include_files.append (full_name)
             if self.follow: self.follow.openInclude ()
             self.info ("OPEN INCL " + full_name)
             self.nextChar ()
             self.expandToken ()
       else :
          self.expandToken ()

   # -----------------------------------------------------------------------

   def nextToken (self) :
       if self.pascal :
          self.simpleToken ()
          while not self.use_input :
             if 0 and self.follow : self.follow.unusedToken ()
             self.simpleToken ()
       else :
          self.expandToken ()
          again = True
          while again :
             again = True
             if self.isToken ('#'):
                self.directive ()
             elif self.isEndOfSource () and len (self.input_stack) != 0 :
                if self.input.isInclude:
                   "check for empty ifndef ... endif include file"
                   if self.file_input.guard == 3 :
                      key = os.path.abspath (self.getFileName ())
                      value = self.file_input.guard_symbol
                      self.guard_map [key] = value
                      # print ("GUARD", key, value)
                   "close include file"
                   self.info ("CLOSE INCL")
                   if self.follow: self.follow.closeInclude ()
                if self.input.isMacro:
                   "close macro"
                   if self.follow: self.follow.closeExpand ()
                   self.active_macros [self.input.macro_name] = False
                self.input = self.input_stack.pop ()
                self.file_input = self.getFileInput ()
                self.repeatChar ()
                self.expandToken ()
             else :
                if not self.use_input :
                   if 0 and self.follow : self.follow.unusedToken ()
                   # print ("unused token:", self.tokenText)
                   self.simpleToken ()
                else :
                   "valid token"
                   again = False
       if self.file_input.guard == 1 or self.file_input.guard == 3 :
          "check for empty ifndef ... endif include file"
          self.file_input.guard = 0
       if self.token == self.identifier :
          "convert identifier to keyword"
          self.lookupKeyword ()
       if self.follow : self.follow.oneToken ()
       if self.debug : print ("TOKEN", self.tokenText)

   # -----------------------------------------------------------------------

   def isEndOfSource (self) :
       return self.token == self.eos

   def isIdentifier (self) :
       return self.token == self.identifier

   def isNumber (self) :
       return self.token == self.number

   def isReal (self) :
       return self.token == self.real_number

   def isString (self) :
       return self.token == self.string_literal

   def isCharacter (self) :
       return self.token == self.character_literal

   def isSeparator (self, value) :
       return self.token == self.separator and self.tokenText == value

   def isKeyword (self, value) :
       return self.token == self.identifier and self.tokenText == value

   # -----------------------------------------------------------------------

   def readIdentifier (self, msg = "Identifier expected") :
       if not self.isIdentifier () :
          self.error (msg)
       value = self.tokenText
       self.nextToken ()
       return value

   def readNumber (self) :
       if not self.isNumber () :
          self.error ("Number expected")
       value = self.tokenText
       self.nextToken ()
       return value

   def readReal (self) :
       if not self.isReal () :
          self.error ("Real number expected")
       value = self.tokenText
       self.nextToken ()
       return value

   def readString (self) :
       if not self.isString () :
          self.error ("String literal expected")
       value = self.tokenText
       self.nextToken ()
       return value

   def readCharacter (self) :
       if not self.isString () :
          self.error ("Character literal expected")
       value = self.tokenText
       self.nextToken ()
       return value

   # -----------------------------------------------------------------------

   def checkSeparator (self, value) :
       if not self.isSeparator (value) :
          self.error (value + " expected")
       self.nextToken ()

   def checkKeyword (self, value) :
       if not self.isKeyword (value) :
          self.error (value + " expected")
       self.nextToken ()

   # -----------------------------------------------------------------------

   def check (self, value) :
       if self.tokenText != value :
          self.error (value + " expected")
       self.nextToken ()

   def tokenToString (self, value) :
       return "token " + str (value)

   # -----------------------------------------------------------------------

   # editor support

   def setSupport (self, support) :
       self.support = support

   def openRegion (self, obj) :
       if self.support != None :
          self.support.openRegion (obj)

   def closeRegion (self) :
       if self.support != None :
          self.support.closeRegion ()

   def openCompletion (self, obj) :
       if self.support != None :
          self.support.openCompletion (obj)

   def closeCompletion (self, obj) :
       if self.support != None :
          self.support.closeCompletion (obj)

   def markDefn (self, obj) :
       if self.support != None :
          self.support.markDefn (obj)

   def markUsage (self, obj) :
       if self.support != None :
          self.support.markUsage (obj)

   def markDefnWithPosition (self, obj, begin_pos, end_pos) :
       if self.support != None :
          self.support.markDefnWithPosition (obj, begin_pos, end_pos)

   def markOutline (self, obj) :
       if self.support != None :
          self.support.markOutline (obj)

   def setInk (self, ink) :
       if self.support != None :
           self.support.setInk (ink)

   def setPaper (self, paper, prev = False) :
       if self.support != None :
          self.support.setPaper (paper)

   def addToolTip (self, tooltip, prev = False) :
       if self.support != None :
          self.support.addToolTip (tooltip)

# --------------------------------------------------------------------------

class FollowClass :
   def start (self, lexer) :
       pass

   def oneCharacter (self) :
       pass
   def whiteSpace (self) :
       pass
   def startComment (self, delta) :
       pass
   def comment (self) :
       pass

   def unusedToken (self) :
       pass
   def oneToken (self) :
       pass

   def startDirective (self) :
       pass
   def endDirective (self) :
       pass
   def unusedDirective (self) :
       pass

   def ifDirective (self) :
       pass
   def ifdefDirective (self) :
       pass
   def ifndefDirective (self) :
       pass
   def elifDirective (self) :
       pass
   def elseDirective (self) :
       pass
   def endifDirective (self) :
       pass
   def defineDirective (self, defn) :
       pass
   def undefDirective (self, name) :
       pass
   def warningDirective (self, text) :
       pass
   def errorDirective (self, text) :
       pass
   def evalDirective (self, text) :
       pass
   def printDirective (self, text) :
       pass

   def openInclude (self) :
       pass
   def closeInclude (self) :
       pass

   def expandMacro (self, defn) :
       pass
   def closeExpand (self) :
       pass
   def openParameters (self) :
       pass
   def closeParameters (self) :
       pass
   def openParam (self, param) :
       pass
   def paramToken (self) :
       pass
   def closeParam (self) :
       pass

# --------------------------------------------------------------------------

# fileNames = [ "file with index 1", "file with index 2" ]

# lexer_keywords = { "if" : token_code_for_if ,
#                    "while" : token_code_for_while }

# lexer_separators = { "<" : Separators (value = code_for_less,
#                                        items = { "=" : ... , "<" : ... }),
#                      ";" : Separators (value = code_for_semicolons,
#                                        items = { }) }

# --------------------------------------------------------------------------

if __name__ == "__main__" :
    lexer = Lexer ()
    lexer.openFile ("pas.g")

    if 0 :
       for i in range (4) :
          print (lexer.tokenText)
          lexer.nextToken ()

       print ("---")

       point = lexer.mark ()
       for i in range (4) :
          print (lexer.tokenText)
          lexer.nextToken ()

       print ("---")
       lexer.rewind (point)
       for i in range (4) :
          print (lexer.tokenText)
          lexer.nextToken ()

    while not lexer.isEndOfSource () :
       print (lexer.tokenText, end = " ")
       lexer.nextToken ()

    point = lexer.mark ()
    lexer.rewind (point)

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
